/**
 * Jackens’ Collection of useful JavaScript helpers. <https://jackens.github.io/jcQuery/>
 */

/**
 * A helper that checks if the given argument is of a certain type.
 *
 * @template T
 *
 * @type {{
 *   (type: BigIntConstructor,  arg: any): arg is bigint;
 *   (type: BooleanConstructor, arg: any): arg is boolean;
 *   (type: NumberConstructor,  arg: any): arg is number;
 *   (type: StringConstructor,  arg: any): arg is string;
 *   (type: SymbolConstructor,  arg: any): arg is symbol;
 *   (type: undefined,          arg: any): arg is undefined | null;
 *   <T extends abstract new (...args: any) => any>(type: T, arg: any): arg is InstanceType<T>;
 * }}
 *
 * @returns {arg is bigint | boolean | number | string | symbol | undefined | null | InstanceType<T>}
 */
export const $is = (/** @type {T} */ type, /** @type {any} */ arg) => arg?.constructor === type

/**
 * Replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.
 */
export const $in = (/** @type {any} */ key, /** @type {any} */ ref) =>
  ($is(String, key) || $is(Number, key) || $is(Symbol, key)) && Object.hasOwnProperty.call(ref ?? Object, key)

/**
 * A helper that checks equality of the given arguments.
 *
 * @returns {boolean}
 */
export const $eq = (/** @type {any} */ x, /** @type {any} */ y) => {
  if (x === y) {
    return true
  }

  const xConstructor = x?.constructor

  if (xConstructor === y?.constructor) {
    if (xConstructor === Number) {
      return (isNaN(x) && isNaN(y)) || +x === +y
    }

    if (xConstructor === Date) {
      return +x === +y
    }

    if (xConstructor === String || xConstructor === RegExp) {
      return '' + x === '' + y
    }

    if (xConstructor === Array) {
      return x.length === y.length && x.every((/** @type {any} */ item, /** @type {number} */ index) => $eq(item, y[index]))
    }

    if (xConstructor === Object) {
      const keysOfX = Object.keys(x)
      return keysOfX.length === Object.keys(y).length && keysOfX.every(key => $eq(x[key], y[key]))
    }

    if (xConstructor === Set || xConstructor === Map) {
      if (x.size !== y.size) {
        return false
      }
      const xa = [...x]
      const ya = [...y]
      return xa.every(xv => ya.find(yv => $eq(xv, yv)))
    }
  }

  return false
}

/**
 * @typedef {keyof HTMLElementTagNameMap | HTMLElement | ShadowRoot} $HTMLElement
 */

/**
 * @typedef {Record<string, any>} $AttributesAndProperties
 */

/**
 * @typedef {$HTMLElement | Text | SVGElement | string | number | $HTMLConfig} $HTMLItem
 */

/**
 * @typedef {[$HTMLElement, $AttributesAndProperties, ...$HTMLItem[]]} $HTMLConfig
 */

/**
 * A lightweight ExtJS-like helper for creating and modifying `HTMLElement`s.
 *
 * - `e`: the modified `HTMLElement` or tag of the element being created.
 * - `attributesAndProperties`: map of attributes and properties. Keys starting with `$` are treated as *properties* to be set on the element being created or modified (without the leading `$`). All other keys are treated as *attributes* to be set by `setAttribute`. Attributes equal to `false` cause the attribute to be removed by `removeAttribute`.
 * - `items`: items to be appended to the element being created or modified.
 *
 * @type {{
 *   <T extends keyof HTMLElementTagNameMap>(e: T, keys?: $AttributesAndProperties, ...items: $HTMLItem[]): HTMLElementTagNameMap[T];
 *   <E extends HTMLElement>(e: E, keys?: $AttributesAndProperties, ...items: $HTMLItem[]): E;
 *   (e: $HTMLElement, keys?: $AttributesAndProperties, ...items: $HTMLItem[]): HTMLElement | ShadowRoot;
 * }}
 *
 * @returns {HTMLElement | ShadowRoot}
 */
export const $h = (
  /** @type {$HTMLElement}             */ e,
  /** @type {$AttributesAndProperties} */ attributesAndProperties = {},
  /** @type {$HTMLItem[]}              */ ...items
) => {
  e = $is(String, e) ? document.createElement(e) : e

  if (e instanceof HTMLElement) {
    for (const name in attributesAndProperties) {
      const value = attributesAndProperties[name]
      if (name[0] === '$') {
        const name1 = name.slice(1)
        if (name1 in e && $is(Object, value)) { // warning: $in cannot be used here!
          for (const name2 in value) {
            // @ts-expect-error:
            e[name1][name2] = value[name2]
          }
        } else {
          // @ts-expect-error:
          e[name1] = value
        }
      } else {
        if (value === true) {
          e.setAttribute(name, '')
        } else if (value === false) {
          e.removeAttribute(name)
        } else {
          e.setAttribute(name, value)
        }
      }
    }
  }

  for (const item of items) {
    e.appendChild(item instanceof HTMLElement || item instanceof SVGElement || item instanceof Text || item instanceof ShadowRoot
      ? item
      : $is(String, item) || $is(Number, item)
        // @ts-expect-error:
        ? document.createTextNode(item)
        : $h(...item))
  }

  return e
}

/**
 * @typedef {keyof SVGElementTagNameMap | SVGElement} $SVGElement
 */

/**
 * @typedef {$SVGElement | Text | string | number | $SVGConfig} $SVGItem
 */

/**
 * @typedef {[$SVGElement, $AttributesAndProperties, ...$SVGItem[]]} $SVGConfig
 */

const /** @type {Record<string, string>} */ NS = {
  xlink: 'http://www.w3.org/1999/xlink'
}

/**
 * A lightweight ExtJS-like helper for creating and modifying `SVGElement`s (conf. `$h`).
 *
 * - `e`: the modified `SVGElement` or tag of the element being created.
 * - `attributesAndProperties`: map of attributes and properties. Keys starting with `$` are treated as *properties* to be set on the element being created or modified (without the leading `$`). All other keys are treated as *attributes* to be set by `setAttribute` or `setAttributeNS`. Attributes equal to `false` cause the attribute to be removed by `removeAttribute` or `removeAttributeNS`.
 * - `items`: items to be appended to the element being created or modified.
 *
 * @type {{
 *   <T extends keyof SVGElementTagNameMap>(e: T, keys?: $AttributesAndProperties, ...items: $SVGItem[]): SVGElementTagNameMap[T];
 *   <E extends SVGElement>(e: E, keys?: $AttributesAndProperties, ...items: $SVGItem[]): E;
 *   (e: $SVGElement, keys?: $AttributesAndProperties, ...items: $SVGItem[]): SVGElement;
 * }}
 *
 * @returns {SVGElement}
 */
export const $s = (
  /** @type {$SVGElement}              */ e,
  /** @type {$AttributesAndProperties} */ attributesAndProperties = {},
  /** @type {$SVGItem[]}               */ ...items
) => {
  e = $is(String, e) ? document.createElementNS('http://www.w3.org/2000/svg', e) : e

  for (const name in attributesAndProperties) {
    const value = attributesAndProperties[name]

    if (name[0] === '$') {
      const name1 = name.slice(1)
      if (name1 in e && $is(Object, value)) { // warning: $in cannot be used here!
        for (const name2 in value) {
          // @ts-expect-error:
          e[name1][name2] = value[name2]
        }
      } else {
        // @ts-expect-error:
        e[name1] = value
      }
    } else {
      const indexOfColon = name.indexOf(':')

      if (indexOfColon >= 0) {
        const /** @type {string=} */ ns = NS[name.slice(0, indexOfColon)]

        if (ns != null) {
          const basename = name.slice(indexOfColon + 1)

          if (value === true) {
            e.setAttributeNS(ns, basename, '')
          } else if (value === false) {
            e.removeAttributeNS(ns, basename)
          } else {
            e.setAttributeNS(ns, basename, $is(String, value) ? value : '' + value)
          }
        }
      } else {
        if (value === true) {
          e.setAttribute(name, '')
        } else if (value === false) {
          e.removeAttribute(name)
        } else {
          e.setAttribute(name, $is(String, value) ? value : '' + value)
        }
      }
    }
  }

  for (const item of items) {
    e.appendChild(item instanceof SVGElement || item instanceof Text
      ? item
      : $is(String, item) || $is(Number, item)
        // @ts-expect-error:
        ? document.createTextNode(item)
        : $s(...item))
  }

  return e
}

/**
 * @typedef {{
 *   [attributeOrSelector: string]: string | number | $CSSNode;
 * }} $CSSNode
 */

/**
 * @typedef {{
 *   [selector: string]: $CSSNode;
 * }} $CSSRoot
 */

const _c = (
  /** @type {$CSSNode} */ node,
  /** @type {string}   */ prefix,
  /** @type {string[]} */ result,
  /** @type {(text: string) => string} */ split
) => {
  const /** @type {[$CSSNode | string[], string][]} */ queue = [[node, prefix]]

  while (queue.length) {
    const [style2, prefix2] = queue.shift() ?? []

    if (style2 == null || prefix2 == null) {
      continue
    }

    if ($is(Array, style2)) {
      result.push(prefix2, prefix2 !== '' ? '{' : '', style2.join(';'), prefix2 !== '' ? '}' : '')
    } else {
      const /** @type {[$CSSNode | string[], string][]} */ todo = []
      let /** @type {string[]} */ attributes = []
      let attributesPushed = false

      for (const key in style2) {
        const value = style2[key]

        if ($is(String, value) || $is(Number, value)) {
          if (!attributesPushed) {
            attributesPushed = true
            attributes = []
            todo.push([attributes, prefix2])
          }

          attributes.push(`${split(key).replace(/([A-Z])/g, (_, letter) => '-' + letter.toLowerCase())}:${value}`)
        } else {
          attributesPushed = false

          const /** @type {string[]} */ newPrefix = []
          const keySplitted = key.split(',')

          for (const prefixItem of prefix2.split(',')) {
            for (const keyItem of keySplitted) {
              newPrefix.push(prefixItem + keyItem)
            }
          }

          todo.push([value, newPrefix.join(',')])
        }
      }

      queue.unshift(...todo)
    }
  }
}

/**
 * CSS-in-JS helper based on the `$CSSRoot` format.
 * An iterative implementation that does not cause a stack overflow exception.
 *
 * The `root` parameter provides a hierarchical description of CSS rules.
 *
 * - Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
 * - All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
 * - In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
 * - Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
 * - Top-level keys that begin with `@` are not concatenated with sub-object keys.
 */
export const $c = (/** @type {$CSSRoot} */ root, /** @type {string} */ splitter = '$$') => {
  const split = (/** @type {string} */ text) => text.split(splitter)[0]
  const /** @type {string[]} */ result = []

  for (const key in root) {
    const value = root[key]

    if (key[0] === '@') {
      result.push(split(key) + '{')
      _c(value, '', result, split)
      result.push('}')
    } else {
      _c(value, split(key), result, split)
    }
  }

  return result.join('')
}

/**
 * A generic helper for syntax highlighting.
 *
 * @returns {$HTMLItem[]}
 */
export const $highlight = (
  /** @type {RegExp} */ pattern,
  /** @type {((chunk: string, index: number) => $HTMLItem)[]} */ highlighters,
  /** @type {string} */ innerText
) => {
  const /** @type {$HTMLItem[]} */ result = []

  innerText.split(pattern).forEach((chunk, index) => {
    index %= highlighters.length
    if (chunk != null) {
      result.push(highlighters[index](chunk, index))
    }
  })

  return result
}

/** @returns {string} */
const _locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ language,
  /** @type {string} */ text,
  /** @type {string} */ version
) => {
  // @ts-expect-error:
  const v = locales?.[language]?.[version]?.[text]
  if ($is(String, v)) {
    return v
  }
  const t = locales?.[language]?.[text]
  return $is(String, t) ? t : text
}

/**
 * Language translations helper.
 */
export const $locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ defaultLanguage,
  /** @type {readonly string[]} */ languages = navigator.languages
) => {
  for (const language of languages) {
    if ($in(language, locales)) {
      // @ts-expect-error:
      return _locale.bind(0, locales, language)
    }
  }

  // @ts-expect-error:
  return _locale.bind(0, locales, defaultLanguage)
}
